/**
 * AsynchronousNodeTalker.java
 * Created on 2005/05/18
 * Copyright (c) by es , All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND 
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF 
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package izayoi.node;

/**
 * 񓯊ʐMNX
 * @author es
 */
public class AsynchronousNodeTalker {

    private class ThreadHolder {
        private Thread thread;
        ThreadHolder(Thread t) {
            thread = t;
        }
        synchronized void clear() {
            thread = null;
        }
        synchronized Thread get() {
            return thread;
        }
    }

    protected AsynchronousNode node;
    protected TalkingQueryChain chain;

    private TalkingEvent _result;
    private ThreadHolder _thread;

    /*
     * spXbh
     */
    private AsynchronousNodeTalker() {
        final Runnable runner = new Runnable() {
            public void run() {
                try {
                    setResult(executeAsyncTalk());
                } finally {
                    fire(getResult());
                    _thread.clear();
                }
            }
        };
        _thread = new ThreadHolder(new Thread(runner));
    }

    /**
     * RXgN^
     * @param node Ăь
     * @param chain ꊇsNG
     */
    public AsynchronousNodeTalker(AsynchronousNode node, TalkingQueryChain chain) {
        this();
        this.node = node;
        this.chain = chain;
    }

    /**
     * RXgN^
     * @param node Ăь
     * @param target ʐMNG
     */
    public AsynchronousNodeTalker(AsynchronousNode node, TalkingQuery target) {
        this();
        this.node = node;
        this.chain = new TalkingQueryChain(target);
    }

    /**
     * RXgN^
     * @param node Ăь
     * @param target ʐMNG
     */
    public AsynchronousNodeTalker(AsynchronousNode node, TalkingQuery[] target) {
        this();
        this.node = node;
        this.chain = new TalkingQueryChain(target);
    }

    /**
     * 񓯊talksꂽƂ̏
     * @return sʂCxg
     * @throws IOException
     */
    protected TalkingEvent executeAsyncTalk() {
        return chain.invoke(node);
    }

    /**
     * ĂьɃCxgʒm
     * @param event
     */
    private void fire(TalkingEvent event) {
        if(node != null) {
            node.fire(event);
        }
    }

    /**
     * sʎ擾
     * @return
     */
    synchronized protected TalkingEvent getResult() {
        return _result;
    }

    /**
     * f
     */
    public final void interrupt() {
        Thread t = _thread.get();
        if (t != null) {
            t.interrupt();
        }
        _thread.clear();
    }

    synchronized private void setResult(TalkingEvent event) {
        _result = event;
    }

    /**
     * 񓯊talk
     */
    public final void talk() {
        Thread t = _thread.get();
        if (t != null) {
            t.start();
        }
    }
}
