package izayoi.servlet;

import izayoi.Constants;
import izayoi.Util;
import izayoi.cache.CacheEntry;
import izayoi.cache.Record;
import izayoi.exception.InvalidParameter;
import izayoi.view.TemplateContext;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
 * delete.cgi
 * @author es
 */
public class DeleteServlet extends BasementServlet {

    /* ( Javadoc)
     * @see izayoi.servlet.BasementServlet#execute(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
     */
    protected void execute(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {

        Map delconf = (Map) request.getSession().getAttribute(COMMAND_DELETE);
        if (delconf == null && !request.getMethod().equalsIgnoreCase("post")) {
            response.sendError(HttpServletResponse.SC_BAD_REQUEST);
            return;
        }

        String[] reqline = this.parseURL(request.getRequestURI());
        if (reqline == null || reqline[0] == null || reqline[0].equals("")) {
            response.sendError(HttpServletResponse.SC_NOT_FOUND);
            return;
        }

        if (delconf == null || request.getMethod().equalsIgnoreCase("get") || request.getParameter(PARAM_RECORD) != null) {
            deleleConf(request, response, Util.urlDecode(reqline[0]));
        } else {
            delete(request, response, delconf);
        }
    }

    /**
     * mF
     * @param request
     * @param response
     * @param target
     */
    private void deleleConf(HttpServletRequest req, HttpServletResponse res, String target) throws ServletException, IOException {

        String type = req.getParameter(PARAM_TYPE);
        String timeID = req.getParameter(PARAM_RECORD);
        if (type == null || timeID == null) {
            throw new InvalidParameter(type == null ? PARAM_TYPE : PARAM_RECORD, "l܂");
        }
        CacheEntry c = cache.getEntry(type, target);
        if (c == null) {
            res.sendError(HttpServletResponse.SC_NOT_FOUND);
            return;
        }
        String tmp[] = timeID.split("_");
        if (tmp.length != 2) {
            throw new InvalidParameter(PARAM_RECORD, "sł");
        }
        Record rec = c.getRecord(Long.parseLong(tmp[0]), tmp[1]);
        if (rec == null) {
            res.sendError(HttpServletResponse.SC_NOT_FOUND);
            return;
        }

        if (rec.containsField("sign")) {
            throw new InvalidParameter(PARAM_RECORD, "̃R[h͏Ă܂");
        }

        Map delconf = new HashMap();
        delconf.put(PARAM_FILE, c.getFileName());
        delconf.put(PARAM_BASEID, rec.getID());
        delconf.put(PARAM_BASETIME, new Long(rec.getTimeStamp()));
        req.getSession().setAttribute(COMMAND_DELETE, delconf);

        TemplateContext ctx = new TemplateContext("delete.txt", req, res);
        String str = rec.getField(Constants.FIELD_NAME_BODY);
        if (str == null) {
            str = rec.getField(Constants.FIELD_NAME_LINK);
        }
		ctx.put("records", str);
		ctx.put("id", rec.getID().substring(0,8));
		ctx.put("time", Util.parseDateString(rec.getTimeStamp()));
        ctx.finish();
    }

    /**
     * 폜s
     * @param request
     * @param response
     * @param delconf
     */
    private void delete(HttpServletRequest req, HttpServletResponse res, Map delconf) throws IOException {
        String filename = (String) delconf.get(PARAM_FILE);
        String id = (String) delconf.get(PARAM_BASEID);
        long time = ((Long) delconf.get(PARAM_BASETIME)).longValue();

        CacheEntry c = cache.getEntry(filename);
        Record r = c.getRecord(time, id);
        c.delete(r);
        req.getSession().removeAttribute(COMMAND_DELETE);

        //@TODO [g߂
        String ret = "/" + Util.getTypeExecutor(c.getType()) + "/" + Util.urlEncode(c.getBaseName());
        res.sendRedirect(ret);
    }

}
