package izayoi.servlet;

import izayoi.Izayoi;
import izayoi.cache.CacheEntry;
import izayoi.cache.Record;

import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
 * hserver.cgih
 * @author es
 */
public class ServerServlet extends BasementServlet {

    private static final Pattern urlPtn = Pattern.compile("\\/[\\w|\\.]+(\\/(\\w+)(\\/(.+)?)?)?");
    private static final String schme = "http://";

    protected void execute(HttpServletRequest req, HttpServletResponse res) throws ServletException, IOException {

        Matcher m = urlPtn.matcher(req.getRequestURI());
        if (!m.find()) {
            res.sendError(HttpServletResponse.SC_BAD_REQUEST);
            return;
        }
        String cmd = (m.group(2) == null) ? "/" : m.group(2);
        String argument = m.group(4);

        res.setContentType("text/plain; charset=UTF-8");
        if (cmd.equals("ping")) {
            ping(req, res);
        } else if (cmd.equals("join")) {
            join(res, argument);
        } else if (cmd.equals("node")) {
            node(res);
        } else if (cmd.equals("bye")) {
            bye(res, argument);
        } else if (cmd.equals("have")) {
            have(res, argument);
        } else if (cmd.equals("get")) {
            get(res, argument);
        } else if (cmd.equals("head")) {
            head(res, argument);
        } else if (cmd.equals("update")) {
            update(res, argument);
        } else if (cmd.equals("/")) {
            root(res);
        }

    }

    /*
     * "ping"̎M
     * @param res
     */
    private void ping(HttpServletRequest req, HttpServletResponse res) throws IOException {
        log("Received ping");
        PrintWriter w = res.getWriter();
        w.println("PONG");
        w.println(req.getRemoteAddr());
    }

    /*
     * "join"̎M
     * @param res
     * @param argument
     */
    private void join(HttpServletResponse res, String argument) throws IOException {
        log("Received join :" + argument);
        argument = argument.replaceAll("\\+", "/");
        URL url = new URL(schme + argument);

        PrintWriter w = res.getWriter();
        if (node.checkExist(url)) {
            w.println("WELCOME");
            node.addNode(url);
        } else {
            w.println("GET OUT");
        }
    }

    /*
     * "node"̎M
     * @param res
     */
    private void node(HttpServletResponse res) throws IOException {
        URL url = node.getRandomNode().getServantURL();
        String s = url.getHost() + ":" + url.getPort() + url.getFile();
        PrintWriter w = res.getWriter();
        w.println(s);
    }

    /*
     * "bye"̎M
     * @param res
     * @param argument
     */
    private void bye(HttpServletResponse res, String argument) throws IOException {
        log("Received bye :" + argument);
        PrintWriter w = res.getWriter();
        argument = argument.replaceAll("\\+", "/");
        URL url = new URL(schme + argument);

        w.println("BYEBYE");
        node.removeNode(url);
    }

    /*
     * "have"̎M
     * @param res
     * @param argument
     */
    private void have(HttpServletResponse res, String argument) throws IOException {
        log("Received have :" + argument);
        PrintWriter w = res.getWriter();
        boolean ret;
        try {
            CacheEntry e = cache.getEntry(argument);
            ret = (e != null);
        } catch (Exception e) {
            ret = false;
        }
        w.println(ret ? "YES" : "NO");
    }

    /*
     * "get"̎M
     * @param res
     * @param argument
     */
    private void get(HttpServletResponse res, String argument) throws IOException {
        log("Received get :" + argument);
        String filename = argument.substring(0, argument.indexOf("/"));
        String time = argument.substring(argument.indexOf("/") + 1);
        CacheEntry ent;
        try {
            ent = cache.getEntry(filename);
            if (ent == null) {
                return;
            }
            PrintWriter w = res.getWriter();
            Record[] ret = (time != null && !time.equals("")) ? ent.getRecords(time) : ent.getRecords();
            for (int i = 0; i < ret.length; i++) {
                w.println(ret[i].parseString());
            }
        } catch (Exception e) {
            e.printStackTrace();
            return;
        }
    }

    /*
     * "head"̎M
     * @param res
     * @param argument
     */
    private void head(HttpServletResponse res, String argument) {
        log("Received head :" + argument);
        String filename = argument.substring(0, argument.indexOf("/"));
        String time = argument.substring(argument.indexOf("/") + 1);
        CacheEntry ent;
        try {
            ent = cache.getEntry(filename);
            if (ent == null) {
                return;
            }
            if (time == null || time.equals("")) {
                time = "0-";
            }
            String[] ret = ent.head(time);

            PrintWriter w = res.getWriter();
            for (int i = 0; i < ret.length; i++) {
                w.println(ret[i]);
            }
        } catch (Exception e) {
            e.printStackTrace();
            return;
        }
    }

    /*
     * hupdateh̎M
     * @param res
     * @param argument
     */
    private void update(HttpServletResponse res, String argument) throws IOException {
        log("Received update :" + argument);
        try {
            String[] arg = argument.split("/");
            node.sendUpdate(null, arg[0], Long.parseLong(arg[1]), arg[2]);
            node.syncUpdate(arg[0], Long.parseLong(arg[1]), arg[2]);
        } catch (Exception e) {
            if (e instanceof IOException) {
                throw (IOException) e;
            } else if (e instanceof ArrayIndexOutOfBoundsException || e instanceof NumberFormatException) {
                res.sendError(HttpServletResponse.SC_BAD_REQUEST);
                return;
            } else {
                e.printStackTrace();
            }
        }
    }

    /*
     * "/"̎M
     * @param res
     */
    private void root(HttpServletResponse res) throws IOException {
        log("Received root");
        PrintWriter w = res.getWriter();
        w.println(Izayoi.VERSION);
        w.println("NodeController : " + node.toString());
        w.println("CacheController : " + cache.toString());
        w.println("Number of nodes : " + node.getAllNodes().size());
    }

}
