/**
 * ViewList.java
 * Created on 2005/06/08
 * Copyright (c) by es , All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND 
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF 
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package izayoi.view;

import izayoi.Constants;
import izayoi.Util;
import izayoi.cache.CacheEntry;
import izayoi.cache.CacheFactory;
import izayoi.cache.Record;

import java.io.IOException;
import java.text.MessageFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;
import java.util.List;

/**
 * listñev[gɗRei
 * @author es
 */
public class ViewList implements DataContainer {

    private static final String ptn = "/{0}/{1}";

    private int count = 0;
    Date lastModified;
    private String id;
    private String name;
	private String size;
	private long fileLength;
    private long time;
    private String type;
    private String URL;
    private String typename;
	private boolean deleted;

    /**
     * CacheEntry[]ev[gp̃Reiz
     * @param entris
     * @return
     */
    public static DataContainer[] parseContent(CacheEntry[] entris) {
        if (entris == null) {
            entris = new CacheEntry[0];
        }

        List l = new ArrayList();
        for (int i = 0; i < entris.length; i++) {
            try {
				DataContainer v =  parseList(entris[i]);
                if (v != null) {
                    l.add(v);
                }
            } catch (IOException e) {
                continue;
            }
        }
		DataContainer[] val = new DataContainer[l.size()];
        val = (DataContainer[]) l.toArray(val);
        return val;
    }

    /**
     * Record[]ev[gp̃Reiz
     * @param record
     * @param sortkey \[gL[
     * @param order \[g̃I[_[
     * @return
     */
    public static DataContainer[] parseContent(Record[] record, String sortkey, boolean order) {
        if (record == null) {
            record = new Record[0];
        }

		//@TODO Xg̃R[hɂ͍ŏIXV烌R[hĂȂ̂łŃ\[g邵Ȃc@
        Comparator sorter = null;
        if(sortkey != null) {
            sorter = new ListSorter(sortkey, order);
        }

        List l = new ArrayList();
        for (int i = 0; i < record.length; i++) {
            try {
				DataContainer v = parseList(record[i]);
                if (v != null) {
                    l.add(v);
                }
            } catch (IOException e) {
                continue;
            }
        }
		DataContainer[] val = new DataContainer[l.size()];
        val = (DataContainer[]) l.toArray(val);
        if(sorter != null) {
            Arrays.sort(val, sorter);
        }
        return val;
    }

    private static DataContainer parseList(CacheEntry entry) throws IOException {
        if (entry == null) {
            return null;
        }
        ViewList v = new ViewList();
        v.id = null;
        v.time = entry.getLastModified();
        v.lastModified = new Date(v.time * 1000);
        v.type = entry.getType();
        v.typename = Util.getTypeName(v.type);
        v.name = entry.getBaseName();
        v.count = entry.getRecordCount();
        v.fileLength = entry.getSize();
        v.size = Util.parseByteString(v.fileLength);
		v.deleted = false;

        //@TODO [gߑł߂
        v.URL = MessageFormat.format(ptn, new Object[] { Util.getTypeExecutor(v.type), Util.urlEncode(v.name)});
        return v;
    }

    private static DataContainer parseList(Record record) throws IOException {
        if (record == null) {
            return null;
        }
        String type = record.getField(Constants.FIELD_NAME_TYPE);
        String name = record.getField(Constants.FIELD_NAME_LINK);
        CacheEntry entry = null;
        if(!record.isDeletedRecord()) {
			entry = CacheFactory.getInstance().getEntry(type, name);
        }

        ViewList v;
        if (entry != null) {
            v = (ViewList) parseList(entry);
            v.id = record.getID();
            v.time = record.getTimeStamp();
            v.lastModified = new Date(v.time * 1000);
        } else {
            v = new ViewList();
			v.deleted = record.isDeletedRecord();
			if(!v.deleted) {
				v.URL = MessageFormat.format(ptn, new Object[] { Util.getTypeExecutor(type), Util.urlEncode(name)});
				v.type = type;
				v.typename = Util.getTypeName(type);
				v.name = name;
			} else {
				v.name = "폜" + Util.parseDateString(record.getDeletedTime());
			}
            v.id = record.getID();
            v.time = record.getTimeStamp();
            v.lastModified = new Date(v.time * 1000);
            v.size = "0byte";
			v.fileLength = 0;
        }
        return v;
    }

    /**
     * @return
     */
    public int getCount() {
        return count;
    }

    /**
     * @return
     */
    public String getDate() {
        return SimpleDateFormat.getDateTimeInstance().format(lastModified);
    }

    /**
     * @return
     */
    public String getId() {
        return id;
    }

    /**
     * @return
     */
    public String getName() {
        return name;
    }

	/**
	 * @return
	 */
	public String getSize() {
		return size;
	}

	/**
	 * @return
	 */
	public long getLength() {
		return fileLength;
	}

    /**
     * @return
     */
    public long getTime() {
        return time;
    }

    /**
     * @return
     */
    public String getType() {
        return type;
    }

    /**
     * @return
     */
    public String getUrl() {
        return URL;
    }

    /**
     * @return
     */
    public String getTypename() {
        return typename;
    }

    /**
     * @return
     */
    public boolean getDeleted() {
        return deleted;
    }

}

class ListSorter implements Comparator {
    private String key;
    private boolean isAsc;

    /**
     * 
     */
    public ListSorter(String key, boolean isAsc) {
        if (key == null || key.equals("")) {
            throw new IllegalArgumentException("bad key name");
        }
        this.key = key;
        this.isAsc = isAsc;
    }

    /* ( Javadoc)
     * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
     */
    public final int compare(Object arg0, Object arg1) {
        return compare((ViewList) arg0, (ViewList) arg1);
    }

    private int compare(ViewList arg0, ViewList arg1) {
        if (key.equals(Constants.SORT_BYCOUNT)) {
            return compare((float) arg0.getCount(), (float) arg1.getCount());
        } else if (key.equals(Constants.SORT_BYSIZE)) {
            return compare(arg0.getLength(), arg1.getLength());
        } else if (key.equals(Constants.SORT_BYLAST)) {
            return compare(arg0.lastModified, arg1.lastModified);
        } else {
            return compare(arg0.getName(), arg1.getName());
        }
    }

    private int compare(String arg0, String arg1) {
        return isAsc ? arg0.compareTo(arg1) : arg1.compareTo(arg0);
    }

    private int compare(Date arg0, Date arg1) {
        return isAsc ? arg0.compareTo(arg1) : arg1.compareTo(arg0);
    }

    private int compare(float arg0, float arg1) {
        float f0 = isAsc ? arg0 : arg1;
        float f1 = isAsc ? arg1 : arg0;

        if (f0 == f1) {
            return 0;
        } else {
            return f0 > f1 ? 1 : -1;
        }
    }
}
