/*
 * Cache.java
 *
 * Created on 2004/08/26, 17:32
 */

package net.sourceforge.shingetsu.cache;

import java.io.File;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;
import java.util.Vector;
import java.util.Arrays;
import java.util.Comparator;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import net.sourceforge.shingetsu.io.MultipleLineReader;
import net.sourceforge.shingetsu.io.MultipleLineWriter;

import java.io.FileNotFoundException;
import java.io.IOException;

/**
 * ̃NX̓LbVt@C\NXłB̃NX̃IuWFNgʂ
 * ČǂLbVt@C𑀍삷邱Ƃł܂B
 * @author Fuktommy
 * @author stokesia
 */
public class Cache {
    
    /**
     * [h̎wtB[hŁAO\܂B
     */    
    public static final int MODE_SEARCH_BEFORE = 0;
    
    /**
     * [h̎wtB[hŁA\܂B
     */    
    public static final int MODE_SEARCH_AFTER = 1;
    
    private File file = null;
    
    private Comparator searchRecordComparator = null;
    private Comparator containsRecordComparator = null;
    
    private FileChannel fileChannel = null;
    private FileLock fileLock = null;

    private Cache() {
    }
    
    /**
     * ̃NX̃CX^X𐶐܂B
     * @param file LbVt@C\FileIuWFNg
     * @throws FileNotFoundException t@C݂Ȃꍇ
     * @throws IOException LbVƊ֘AÂ`l̏Ɏsꍇ
     * @throws NullPointerException null̏ꍇ
     */
    public Cache(File file) throws FileNotFoundException, IOException, NullPointerException {
        if(!file.exists()) {
            throw new  FileNotFoundException("IuWFNg̕\t@C܂B");
        } 
        if(file == null) {
            throw new NullPointerException("IuWFNgւ̎QƂnullłB");
        }
        
        this.file = file;
        
        searchRecordComparator = new Cache.SearchRecordComparator();
        containsRecordComparator = new Cache.ContainshRecordComparator();
        
        fileChannel = MultipleLineReader.getFileChannel(file);
    }

    /**
     * V̎dlɏ]͂ČsAɈvsԂ܂B
     * @param timeArgument V̎dlɏ]
     * @throws IOException LbVt@C̑쒆ɓo̓G[ꍇ
     * @throws IllegalArgumentException sȏꍇ
     * @throws NumberFormatException ̏sȏꍇ
     * @return Ɉvsō\zB݂Ȃꍇnull
     */    
    public String[] searchRecord(String timeArgument)
        throws NumberFormatException, IOException, IllegalArgumentException {
        String[] records = null;
        
        if(timeArgument.matches("^(\\d+)")) {
            /*  regex pattern = ^(\d+)  */
            long time = Long.parseLong(timeArgument);
            
            String record =searchRecord(time);
            
            if(record == null) {
                records = null;
            } else {
                records = new String[1];
                records[0] = record;
            }
        } else if(timeArgument.matches("^-(\\d+)")) {
            /*  regex pattern = ^-(\d+) */
            String[] tokens = timeArgument.split("-");
            
            long time = Long.parseLong(tokens[1]);
         
            records = searchRecord(time, MODE_SEARCH_BEFORE);
        } else if(timeArgument.matches("^(\\d+)-")) {
            /*  regex pattern = ^(\d+)- */
            String[] tokens = timeArgument.split("-");
            
            long time = Long.parseLong(tokens[0]);
            
            records = searchRecord(time, MODE_SEARCH_AFTER);
        } else if(timeArgument.matches("^(\\d+)-(\\d+)")) {
            /*  regex pattern = ^(\d+)-(\d+) */
            String[] tokens = timeArgument.split("-");
            
            long beginTime = Long.parseLong(tokens[0]);
            long endTime = Long.parseLong(tokens[1]);
            
            records = searchRecord(beginTime, endTime);
        } else if(timeArgument.matches("^(\\d+)/([0-9a-f]+)")) {
            /*  regex pattern = ^(\d+)/([0-9a-f]+)   */
            String[] tokens = timeArgument.split("/");
            
            long time = Long.parseLong(tokens[0]);
            String md5 = tokens[1];
            
            String record = searchRecord(time, md5);
            if(record == null) {
                records = null;
            } else {
                records = new String[1];
                records[0] = record;
            }
        } else {
            throw new IllegalArgumentException("̎w肪słB");
        }
        
        return records;
    }
    
    /**
     * Ŏw肳ꂽ̍s܂B
     * @param time V̎dlɏ]
     * @throws IOException LbVt@C̑쒆ɓo̓G[ꍇ
     * @throws NumberFormatException ̏sȏꍇ
     * @return ɈvsB݂Ȃꍇnull
     */    
    public String searchRecord(long time) throws IOException, NumberFormatException {
        String timeStamp = new Long(time).toString();
        String[] records = MultipleLineReader.read(file);

        Arrays.sort(records, searchRecordComparator);
        
        int index = Arrays.binarySearch(records, timeStamp, searchRecordComparator);

        String record = null;
        
        if(index >= 0) {
            record = records[index];
            String[] tokens = record.split("<>");
            
            if(!timeStamp.equals(tokens[0])) {
                record = null;
            }
        }

        return record;
    }
    
    /**
     * Ŏw肳ꂽ̑O̍sw肳ꂽ[hŌ܂B
     * @param time Jn
     * @param mode [h
     * @throws IOException LbVt@C̑쒆ɓo̓G[ꍇ
     * @throws NumberFormatException ̏sȏꍇ
     * @throws IllegalArgumentException [h̎w肪sȏꍇ
     * @return Ɉvs̔zB݂Ȃꍇnull
     */    
    public String[] searchRecord(long time, int mode)
        throws IOException, NumberFormatException, IllegalArgumentException {
        if(!((mode == MODE_SEARCH_AFTER) || (mode == MODE_SEARCH_BEFORE))) {
            throw new IllegalArgumentException("[h̎w肪słB");
        }
        
        String timeStamp = new Long(time).toString();
        
        String[] records = MultipleLineReader.read(file);
        
        if(records.length == 0) {
            return null;
        }
        
        Arrays.sort(records, searchRecordComparator);
        
        int index = Arrays.binarySearch(records, timeStamp, searchRecordComparator);
        
        assert (index >= 0) || (index < 0);
        
        index = checkBorderCondition(index, records);

        assert (index >= 0) && (index < records.length);
        //System.out.println("index = " + index);

        int begin = 0;
        int end = 0;
        
        if(mode == MODE_SEARCH_BEFORE) {
            begin = 0;
            
            if((index - 1) >= 0) {
                end = --index;
            }
        } else {
            if(timeStamp.equals(records[index].split("<>")[0])) {
                if((index + 1) != records.length) {
                    begin = ++index;
                }
            } else if(timeStamp.equals("0")){
                begin = 0;
            } else {
                begin = index;
            }
            end = records.length - 1;
        }
        
        Vector recordList = new Vector();
                
        for(int i = begin; i <= end; i++) {
                recordList.add(records[i]);
        }
        
        return (toArray(recordList));
    }
    
    /**
     * Ŏw肳ꂽԂ̎̍s܂B
     * @param startTime ͈͂̊Jn
     * @param endTime ͈͂̏I
     * @throws IOException LbVt@C̑쒆ɓo̓G[ꍇ
     * @throws NumberFormatException ̏sȏꍇ
     * @return Ɉvs̔zB݂Ȃꍇnull
     */    
    public String[] searchRecord(long startTime, long endTime) 
        throws IOException, NumberFormatException {
        if(startTime > endTime) {
            long time = startTime;
            startTime = endTime;
            endTime = time;            
        }
        
        String startTimeStamp = new Long(startTime).toString();
        String endTimeStamp = new Long(endTime).toString();
        
        String[] records = MultipleLineReader.read(file);

        Arrays.sort(records, searchRecordComparator);
        
        int start = Arrays.binarySearch(records, startTimeStamp, searchRecordComparator);
        int end = Arrays.binarySearch(records, endTimeStamp, searchRecordComparator);
        
        assert (start < 0) || (start < records.length);
        assert (end < 0) || (end < records.length);
        
        start = checkBorderCondition(start, records);
        end = checkBorderCondition(end, records);
        
        assert (start >= 0) && (start < records.length);
        assert (end >= 0) && (end < records.length);
        
        if(startTimeStamp.equals(records[start].split("<>")[0])) {
            if((start + 1) != records.length) {
                start++;
            }
        }
        
        if((end - 1) >= 0) {
            end--;
        }

        assert (start >= 0) && (start < records.length);
        assert (end >= 0) && (end < records.length);
        
        Vector recordList = new Vector();
        
        for(int i = start; i <= end; i++) {
            recordList.add(records[i]);
        }
        
        return (toArray(recordList));
    }
    
    private int checkBorderCondition(int index, String[] records) {
        if(index < 0) {
            index = Math.abs(index);
            if((index >= records.length) && ((index  - 1) >= 0)) {
                index = records.length -1;
            }
        }
        
        return index;
    }
    
    /**
     * Ŏw肳ꂽƎʎqɈvs܂B
     * @param time 
     * @param md5 ʎq
     * @throws IOException LbVt@C̑쒆ɓo̓G[ꍇ
     * @throws NumberFormatException ̏sȏꍇ
     * @return ɈvsB݂Ȃꍇnull
     */    
    public String searchRecord(long time, String md5) 
        throws IOException, NumberFormatException {
        String timeStamp = new Long(time).toString();
        
        String[] records = MultipleLineReader.read(file);
        Arrays.sort(records, containsRecordComparator);
        
        String key = timeStamp + "<>" + md5;
        
        int index = Arrays.binarySearch(records, key, containsRecordComparator);
        String record = null;
        
        if(index >= 0) {
            record = records[index];
            String[] tokens = record.split("<>");
            
            if(!(timeStamp.equals(tokens[0]) && md5.equals(tokens[1]))) {
                record = null;
            }
        }

        return record;
    }

    /**
     * Ŏw肳ꂽsLbVt@C܂ނǂԂ܂B
     * @param time 
     * @param md5 ʎq
     * @throws IOException LbVt@C̑쒆ɓo̓G[ꍇ
     * @throws NumberFormatException ̏sȏꍇ
     * @return ܂߂true
     */    
    public boolean containsRecord(long time, String md5) 
        throws IOException, NumberFormatException {
        String[] records = MultipleLineReader.read(file);
        
        if(records.length == 0) {
            return false;
        }

        Arrays.sort(records, containsRecordComparator);
        
        String timeStamp = new Long(time).toString();
        String key = timeStamp + "<>" + md5;
        
        int index = Arrays.binarySearch(records, key, containsRecordComparator);
        
        boolean result = false;
        
        if(index >= 0) {
            String[] tokens = records[index].split("<>");
            result = timeStamp.equals(tokens[0]) && md5.equals(tokens[1]);
        }

       return result;
    }
    
    /**
     * LbVt@C̃R[h擾܂B
     * @return R[hStringz
     * @throws FileNotFoundException LbVt@C݂Ȃꍇ
     * @throws IOException o̓G[ꍇ
     */    
    public String[] getRecords() throws FileNotFoundException, IOException {
        String[] records = MultipleLineReader.read(file);
        
        Arrays.sort(records, containsRecordComparator);
         
        return records;
    }
    
    /**
     * LbVt@C̃R[h擾܂B
     * @return R[h
     * @throws FileNotFoundException LbVt@C݂Ȃꍇ
     * @throws IOException o̓G[ꍇ
     */    
    public int getRecordCount() throws FileNotFoundException, IOException {
        String[] records = MultipleLineReader.read(file);

        int count = 0;
        
        if(records != null) {
            count  = records.length;
        }
        
        return count;
    }
    
    /**
     * LbVt@C̍Ō̃R[h擾܂B
     * @return LbVt@C̍Ō̃R[h
     * @throws FileNotFoundException LbVt@C݂Ȃꍇ
     * @throws IOException o̓G[ꍇ
     */    
    public String getLastRecord() throws FileNotFoundException, IOException {
        String[] records = getRecords();

        Arrays.sort(records,searchRecordComparator);
        
        String record = null;
        int index = records.length - 1;
        
        if(index  >= 0) {
            record = records[index];
        }
        
        return record;
    }
    
    /**
     * LbVt@CɃR[hǉ܂B
     * @throws FileNotFoundException LbVt@C݂Ȃꍇ
     * @throws IOException o̓G[ꍇ
     * @param record ǉ郌R[h
     */    
    public void append(String record) throws FileNotFoundException, IOException {
        MultipleLineWriter.append(record, this.file);
    }
    
    /**
     * LbVt@C̃R[h擾܂B
     * @param time 폜郌R[h̃^CX^v
     * @param id 폜郌R[h̎ʎq
     * @throws FileNotFoundException LbVt@C݂Ȃꍇ
     * @throws IOException o̓G[ꍇ
     * @throws NumberFormatException R[h̐̃tH[}bgsȏꍇ
     */    
    public void deleteRecord(long time, String id) 
        throws FileNotFoundException, IOException, NumberFormatException {
//        String[] records = getRecords();
//        
//        Arrays.sort(records, containsRecordComparator);
//        
//        String timeStamp = new Long(time).toString();
//        
//        String key = timeStamp + "<>" + id;
//        int index = Arrays.binarySearch(records, key, containsRecordComparator);
//        
//        int group = 1;
//        
//        if(index >= 0) {
//            String regex = "^(\\d+<>[0-9a-f]+)(.*)";
//            Matcher matcher = Pattern.compile(regex).matcher(records[index]);
//            assert matcher.matches();
//            records[index] = matcher.group(group);
//        }

        //fix
        List l = new ArrayList(Arrays.asList(getRecords()));
        Collections.sort(l, searchRecordComparator);
        String key = String.valueOf(time) + "<>" + id;
        int index = Collections.binarySearch(l, key, containsRecordComparator);
        if(index >= 0) {
            l.remove(index);
            MultipleLineWriter.write((String[])l.toArray(new String[l.size()]), this.file);
        }
    }
    
    /**
     * ŗ^ꂽCacheNX̃IuWFNgǂԂ܂B
     * @param object CacheNX̃IuWFNg
     * @return true
     */    
    public boolean equals(Object object) {
        boolean result = false;
        Cache comparedCache = null;

        try {
            comparedCache = (Cache) object;
        } catch(ClassCastException e) {
            result = false;
        }
        
        if(comparedCache != null) {
            File comparedFile = comparedCache.getFile();
            result = file.equals(comparedFile);
        }
        
        return result;
    }
    
    /**
     * ̃LbVt@CFileNX̃IuWFNg擾܂B
     * @return LbVt@CFileIuWFNg
     */    
    public File getFile() {
        return file;
    }
    
    /**
     * LbṼR[hƎʎqŎɃ\[g܂B
     * @throws FileNotFoundException LbVt@C݂Ȃꍇ
     * @throws IOException o̓G[ꍇ
     */    
    public void sortRecord() throws FileNotFoundException, IOException {
        String[] records = getRecords();
        Arrays.sort(records, containsRecordComparator);
        MultipleLineWriter.write(records, this.file);
    }
    
    public void lock() throws IOException {
        fileLock = fileChannel.lock();
    }
    
    public void unlock() throws IOException {
        fileLock.release();
    }
    
    public void close() throws IOException {
        fileChannel.close();
    }
    
    private String[] toArray(Vector recordList) {
        Enumeration enumeration = recordList.elements();
        String[] records = new String[recordList.size()];
        
        for(int i = 0; enumeration.hasMoreElements(); i++) {
            records[i] = (String) enumeration.nextElement();
        }
        
        return records;
    }
    
    private class SearchRecordComparator implements Comparator {
        
        public int compare(Object obj1, Object obj2) throws NumberFormatException {
            int timeStampGroup = 1;
            
            String record1 = (String) obj1;
            String record2 = (String) obj2;
            
            String regex = "^(\\d+)";
            Pattern pattern = Pattern.compile(regex);
            
            Matcher matcher1 = pattern.matcher(record1);
            Matcher matcher2 = pattern.matcher(record2);
            
            matcher1.find();
            matcher2.find();
            
            String timeStamp1 = matcher1.group(timeStampGroup);
            String timeStamp2 = matcher2.group(timeStampGroup);
            
            long time1 = Long.parseLong(timeStamp1);
            long time2 = Long.parseLong(timeStamp2);
            
            int result = 0;
            
            if(time1 < time2) {
                result = -1;
            } else if(time1 == time2) {
                result = 0;
            } else {
                result = 1;
            }
            
            return result;
        }
        
    }
    
    private class ContainshRecordComparator implements Comparator {
        
        public int compare(Object obj1, Object obj2) throws NumberFormatException {
            int timeStampGroup = 1;
            int idGroup = 2;
            
            String record1 = (String) obj1;
            String record2 = (String) obj2;
            
            String regex = "^(\\d+)<>([0-9a-f]+)";
            Pattern pattern = Pattern.compile(regex);
            
            Matcher matcher1 = pattern.matcher(record1);
            Matcher matcher2 = pattern.matcher(record2);
            
            matcher1.find();
            matcher2.find();
            
            String timeStamp1 = matcher1.group(timeStampGroup);
            String timeStamp2 = matcher2.group(timeStampGroup);
            
            long time1 = Long.parseLong(timeStamp1);
            long time2 = Long.parseLong(timeStamp2);
            
            int result = 0;
            
            if(time1 < time2) {
                result = -1;
            } else if(time1 == time2) {
                String id1 = matcher1.group(idGroup);
                String id2 = matcher2.group(idGroup);
                
                result = id1.compareTo(id2);
            } else {
                result = 1;
            }
            
            return result;
        }
    }
    
}
